#!/usr/bin/python2
from __future__ import print_function
import sys
import os
import psycopg2
from os.path import join
import argparse
from subprocess import PIPE, Popen
from datetime import datetime, timedelta
import traceback

def log2stderr(msg):
    print(msg, file=sys.stderr)

class ArchiveDatabaseQuery(object):
    def __init__(self, conn):
        self.conn = conn
        self.cursor = self.conn.cursor()
        self.query = None
        self.id = None
        self.begin = None
        self.end = None

    def get_query(self):
        self.cursor.execute("select id,query,interval_begin,interval_end from query_queue where started IS NULL limit 1;")
        return self.cursor.fetchone()

    def __enter__(self):
        try:
            q = self.get_query()
            if not q:
                log2stderr("Found no suitable query in queue")
            else:
                self.id, self.query, self.begin, self.end = self.get_query()
                self._query_started(datetime.now())
        except:
            log2stderr("Unhandled error:")
            traceback.print_exc(20, sys.stderr)
        finally:
            return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is None:
            # Only finish query when things went well
            self._query_finished(datetime.now())
        else:
            traceback.print_exception(exc_type, exc_value, exc_traceback, 20, sys.stderr)
            # If something went wrong, reset the query and
            # hope that next attempt goes well
            self._query_started(None)

        self.cursor.close()
        return True

    def _query_started(self, when):
        if self.id:
            self.cursor.execute("update query_queue set started = %(when)s where id = %(id)s;", {"when": when, "id": self.id})
            self.conn.commit()

    def _query_finished(self, when):
        if self.id:
            self.cursor.execute("update query_queue set finished = %(when)s where id = %(id)s;", {"when": when, "id": self.id})
            self.conn.commit()

def datetime_compare(dta, dtb, resolution):
    """
    Calculate sign(dta - dtb) but with limited resolution of datetimes

    returns:
        -1 if dta > dtb
        0 if dta == dtb
        1 if dta < dtb

    Resolution:
        0 = year
        1 = month
        2 = day
        3 = hour
        4 = minute
        5 = second

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2013-01-08 04:00:32"), 0)
    1

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-02-08 04:00:32"), 0)
    0
    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-02-08 04:00:32"), 1)
    -1

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-02 04:00:32"), 1)
    0
    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-02 04:00:32"), 2)
    1

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 01:00:32"), 2)
    0
    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 01:00:32"), 3)
    1

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 02:34:32"), 3)
    0
    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 02:34:32"), 4)
    -1

    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 02:00:47"), 4)
    0
    >>> datetime_compare(archive_datetime("2014-01-03 02:00:32"), archive_datetime("2014-01-03 02:00:47"), 5)
    -1
    """
    diff = 0
    if diff == 0 and resolution >= 0:
        diff = dta.year - dtb.year

    if diff == 0 and resolution >= 1:
        diff = dta.month - dtb.month

    if diff == 0 and resolution >= 2:
        diff = dta.day - dtb.day

    if diff == 0 and resolution >= 3:
        diff = dta.hour - dtb.hour

    if diff == 0 and resolution >= 4:
        diff = dta.minute - dtb.minute

    if diff == 0 and resolution >= 5:
        diff = dta.second - dtb.second

    if diff < 0:
        return -1
    if diff > 0:
        return 1
    return 0


def archives(roots, interval):
    """
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-08 04:00:32"), datetime.now())))
    ['./tests/fixtures/2014_01/08/11.bz2', './tests/fixtures/2014_06_27_11']
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-08 11:00:32"), archive_datetime("2014-01-08 11:00:43"))))
    ['./tests/fixtures/2014_01/08/11.bz2']
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-08 10:00:32"), archive_datetime("2014-01-08 11:00:43"))))
    ['./tests/fixtures/2014_01/08/11.bz2']
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-08 11:00:32"), archive_datetime("2014-01-08 12:00:43"))))
    ['./tests/fixtures/2014_01/08/11.bz2']
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-10 04:00:32"), datetime.now())))
    ['./tests/fixtures/2014_06_27_11']
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-06-27 14:00:32"), datetime.now())))
    []
    >>> list(archives(["./tests/fixtures"], (archive_datetime("2014-01-08 04:00:32"), archive_datetime("2014-01-08 07:00:32"))))
    []
    >>> list(archives(["./tests/fixtures"], (datetime.fromtimestamp(0), datetime.now())))
    ['./tests/fixtures/2013_11/14/08.bz2', './tests/fixtures/2014_01/08/11.bz2', './tests/fixtures/2014_06_27_11']
    >>> list(archives(["./not_a_dir"], (datetime.fromtimestamp(0), datetime.now())))
    []
    """
    within_interval = lambda interval, dt, resolution: datetime_compare(dt, interval[0], resolution) >= 0 and datetime_compare(dt, interval[1], resolution) <= 0

    for root in roots:
        try:
            for path in sorted(os.listdir(root)):
                parts = path.split("_", 3)
                if len(parts) == 4:
                    try:
                        if within_interval(interval, datetime(int(parts[0]), int(parts[1]), int(parts[2]), int(parts[3])), 3):
                            yield join(root, path)
                    except ValueError:
                        continue

                if len(parts) != 2:
                    continue

                try:
                    yearmonth = datetime(int(parts[0]), int(parts[1]), 1)
                except ValueError:
                    # Failed to cast directory part to int
                    continue

                if ((yearmonth.year < interval[0].year)
                        or (yearmonth.year == interval[0].year and yearmonth.month < interval[0].month)
                        or (yearmonth.year > interval[1].year)
                        or (yearmonth.year == interval[1] and yearmonth.month > interval[1].month)):
                    continue

                for day_s in sorted(os.listdir(join(root, path))):
                    try:
                        day = timedelta(days=int(day_s)-1)
                    except ValueError:
                        continue

                    if not within_interval(interval, yearmonth + day, 2):
                        continue

                    for hour_s in os.listdir(join(root, path, day_s)):
                        try:
                            hour = timedelta(hours = int(os.path.splitext(hour_s)[0]))
                        except ValueError:
                            continue

                        if within_interval(interval, yearmonth + day + hour, 3):
                            yield join(root, path, day_s, hour_s)
        except OSError:
            pass


datetime_fmt = "%Y-%m-%d %H:%M:%S"
def archive_datetime(string):
    """
    Note: only used for tests
    >>> archive_datetime("2014-01-19 10:00:01")
    datetime.datetime(2014, 1, 19, 10, 0, 1)
    >>> archive_datetime("Foo-01-19 10:00:01")
    Traceback (most recent call last):
    ArgumentTypeError: Foo-01-19 10:00:01 is not a valid interval endpoint
    """
    try:
        return datetime.strptime(string, datetime_fmt)
    except ValueError:
        msg = "%s is not a valid interval endpoint" % string
        raise argparse.ArgumentTypeError(msg)

def query_archive(archive_path, query):
    # we only use bz2 for compressing files currently
    cattool = {
        '.bz2': 'bzcat',
        '.gz': 'gzcat',
        '.xz': 'xzcat',
        #'nyan': 'nyancat', etc
        '': 'cat',
    }
    (_, ext) = os.path.splitext(archive_path)
    if not cattool.has_key(ext):
        log2stderr("Don't know how to read archive %s: unrecognized extension %s" % (archive_path, ext))
    unpack = Popen([cattool[ext], archive_path], stdout=PIPE, stderr=PIPE)
    logfilter = Popen(["logfilter", "-j", query], stdin = unpack.stdout, stdout=PIPE, stderr=PIPE)
    unpack.stdout.close()
    for block in iter(lambda: logfilter.stdout.read(1 << 20), ""):
        yield block
    bze = unpack.stderr.read()
    if bze:
        log2stderr("Archive query error.")
        log2stderr(str(bze))
        log2stderr("Abandoning query.")
    unpack.stderr.close()

def filter_logs(paths, interval, query):
    for archive in archives(paths, interval):
        for res in query_archive(os.path.abspath(archive), query):
            yield res

def interactive_query(args):
    path = args.path
    if not path:
        path = ["/opt/logger/archive", "/var/log/op5/logger"]
    interval = args.begin, args.end
    query = args.query
    for log in filter_logs(path, interval, query):
        yield log

def do_database_query(path):
    conn = psycopg2.connect("dbname=op5lsu user=op5lsu")
    with ArchiveDatabaseQuery(conn) as dbquery:
        if dbquery.query:
            interval = (dbquery.begin, dbquery.end)
            query = dbquery.query
            for log in filter_logs(path, interval, query):
                yield log
    conn.close()

def database_query(args):
    return do_database_query(args.path)

def read_timestamp(arg):
    return datetime.fromtimestamp(float(arg))

def get_args():
    parser = argparse.ArgumentParser(description="op5 Logger Archive Querier")

    parser.add_argument("--path", "-p", help="Path to op5 Logger archives", default=[], action='append')

    subparsers = parser.add_subparsers(help = "sub-command help")
    parser_database = subparsers.add_parser("database", help="fetch queries from database")
    parser_database.set_defaults(func=database_query)

    parser_interactive = subparsers.add_parser("interactive", help="read query from command line")
    parser_interactive.add_argument("query", help="a logfilter query")
    parser_interactive.add_argument("--begin", "-b", help="Beginning of interval to search", type=read_timestamp, default=datetime.fromtimestamp(0))
    parser_interactive.add_argument("--end", "-e", help="End of interval to search", type=read_timestamp, default=datetime.now())
    parser_interactive.set_defaults(func=interactive_query)
    return parser.parse_args()

def main_func():
    args = get_args()
    matches = args.func(args)
    for row in matches:
        print(row, end="")

if __name__ == '__main__':
    if len(sys.argv) == 1:
        import doctest
        doctest.testmod()
    else:
        main_func()
